# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class Issue:

    def __init__(self, path, message, checker_id,
                 line_number_start=None, line_number_end=None,
                 line_position=None,
                 source_code=None,
                 severity=None, category=None):
        """
        :param path: filename or folder name
        :param message: What is wrong
        :param checker_id: by which checker this was found
        :param line_number_start: where the issue starts in the file
        :param line_number_end: where the issue ends in the file
        (same as start for one liners)
        :param line_position: The char to highlight in line
        :param severity: unilint unified severity
        :param category: checker specific category
        """
        self.path = path
        self.message = message
        self.checker_id = checker_id
        self.line_number_start = line_number_start
        self.line_number_end = line_number_end
        self.line_position = line_position
        self.source_code = source_code
        self.severity = severity
        self.category = category

    def __repr__(self):
        attributes = [self.path,
                      self.message,
                      self.line_number_start,
                      self.line_number_end,
                      self.line_position,
                      self.source_code,
                      self.severity,
                      self.category]
        attributes = [a for a in attributes if a is not None]
        return "[%s]%s"%(self.checker_id, ":".join([str(a) for a in attributes]))



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class UnilintPluginInitException(Exception):
    pass

# pylint: disable-msg=W0613
__pychecker__ = '--unusednames=cls,options,subdirs,issue_list'


class UnilintPlugin(object):
    '''
    Base class for all plugins defines all methods that will get invoked by unilint
    A plugin may just categorize a file or folder, or it may check a file or folder.
    Categorization is done as an extra step so that many plugins can share the
    categorization effort instead of duplicating it. Category IDs are up to the plugins.
    '''

    def __init__(self, shell_function):
        """
        The constructor takes the shell function so that unit tests can inject a mock function

        :param shell_function: a function with the signature of common.run_shell_command
        :raises UnilintPluginInitException: to signal the plugin cannot execute
        """
        self.shell_cmd = shell_function

    @classmethod
    def get_id(cls):
        """
        :returns: short lowercase string
        """
        raise Exception('get_id not implemented by Plugin class')

    def get_meta_information(self):
        """
        :returns: an empty or a newline terminated string with information about the system relevant to this plugin, if any.
        """
        return ''

    @classmethod
    def is_enabled_by_default(cls):
        """
        :returns: bool
        """
        return True

    @classmethod
    def get_depends(cls):
        """
        :returns: None or list of string ids of plugins on which this plugin depends (e.g. for categories)
        """
        return None

    def categorize_type(self, options, path, subdirs, files):
        """
        returns a dict that can contain categories for path and any of files.
        dict key must be full path (os.path.join(path, filename)).
        subdirs can be used for heuristics, but should not be classified, they will be traversed later.

        :param path: the folder to analyze
        :param subdirs: names of direct subdirs
        :param files: names of files in path
        :returns: dict of str->[str], with resourcename (file or folder), mapped to list of category string ids
        """
        return None

    def check_resource(self, options, path, type_categories):
        """
        :returns: list of unilint.Issue or None.
        """
        return None

    def print_formatted(self, options, issue_list):
        """
        :returns :
        """
        raise Exception('format not implemented by Plugin class')



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import re
from unilint.cpp_source_plugin import AbstractCppPlugin, CPP_EXTENSIONS
from unilint.unilint_plugin import UnilintPluginInitException
from unilint.unilint_main import LEVEL_WARNING, LEVEL_STYLE, LEVELS
from unilint.issue import Issue

PROGRESS_PATTERN = re.compile('^[0-9]+/[0-9]')

__pychecker__ = '--unusednames=cls,options,subdirs'


class CppcheckPlugin(AbstractCppPlugin):

    def __init__(self, shell_function):
        super(CppcheckPlugin, self).__init__(shell_function)

    @classmethod
    def get_id(cls):
        return 'cppcheck'

    def get_meta_information(self):
        cmd = "cppcheck --version"
        value, output, message = self.shell_cmd(cmd,
                                                shell=True,
                                                us_env=True,
                                                ignore_returncodes=[127])
        if value == 0:
            return output
        raise UnilintPluginInitException("ERROR Cannot find cppcheck, install via apt or pip\n%s"%message)

    def check_resource(self, options, path, type_categories):
        if not 'cpp-file' in type_categories:
            return None

        levels = []

        if options.level < LEVELS[LEVEL_WARNING]:
            levels.append('information')
            levels.append('missingInclude')
            levels.append('unusedFunction')

        if options.level == LEVELS[LEVEL_STYLE]:
            levels = ['all']

        lvl_cmd = ''
        if levels != []:
            lvl_cmd = "--enable=%s"%','.join(levels)

        template = '{file}:{line}:{id}:{message}'

        if 'cpp-file' in type_categories:
            cmd = "cppcheck --template %s %s %s"%(template, lvl_cmd, path)

        if options.verbose:
            print(cmd)

        _, output, _ = self.shell_cmd(cmd, shell=True, us_env=True, cwd=os.path.dirname(path), capture_std_err=True)

        if options.raw:
            print output
            return None
        else:
            return self.parse(output)

    def parse(self, output):
        ' example output'
        issues = []
        for line in output.splitlines():
            ignore = False
            if line.strip() == '':
                continue
            for prefix in ['Cppcheck - ', "Checking ", "::missingInclude:Cppcheck cannot find all the include file"]:
                if line.startswith(prefix):
                    ignore = True
                    break
            if ignore:
                continue

            if PROGRESS_PATTERN.match(line):
                continue
            (filename, line_number, category, message) = line.split(':', 3)
            # severity=None
            # if category.startswith('E'):
            #     severity = "warning"
            # if category.startswith('W'):
            #     severity = "style"

            issue = Issue(path=filename,
                          message=message.strip(),
                          checker_id=self.get_id(),
                          line_number_start=line_number,
                          line_number_end=line_number,
                          # severity=severity,
                          category=category)
            issues.append(issue)
        return issues



from setuptools import setup

import imp


def get_version():
    ver_file = None
    try:
        ver_file, pathname, description = imp.find_module('__version__', ['src/unilint'])
        vermod = imp.load_module('__version__', ver_file, pathname, description)
        version = vermod.VERSION
        return version
    finally:
        if ver_file is not None:
            ver_file.close()


setup(name='unilint',
      version=get_version(),
      packages=['unilint', 'roslint'],
      package_dir={'': 'src'},
      scripts=["scripts/roslint", "scripts/unilint"],
      install_requires=[],
      author="Thibault Kruse",
      author_email="thibault.kruse@gmx.de",
      url="https://github.com/tkruse/unilint",
      download_url="https://github.com/tkruse/unilint",
      keywords=["lint", "python", "pylint",
                "pychecker", "pyflakes", "pep8",
                "cppcheck", "linux"],
      classifiers=["Programming Language :: Python",
                   "License :: OSI Approved :: BSD License"],
      description="script wrapping static code analyzers producing unified output",
      long_description="""Unilint runs over files or folders and makes calls to static code analysis tools, which are not installed with unilint. Depends on shell tools like gnu find and grep.""",
      license="BSD")



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from unilint.python_source_plugin import AbstractPythonPlugin
from unilint.unilint_plugin import UnilintPluginInitException
from unilint.unilint_main import LEVEL_WARNING, LEVEL_STYLE, LEVELS
from unilint.issue import Issue

import re

NEWLINE_MATCHER=re.compile('^.+:[0-9]+: \[')


class PylintPlugin(AbstractPythonPlugin):

    def __init__(self, shell_function):
        super(PylintPlugin, self).__init__(shell_function)

    @classmethod
    def get_id(cls):
        return 'pylint'

    def get_meta_information(self):
        cmd = "pylint --version"
        value, output, message = self.shell_cmd(cmd,
                                                shell=True,
                                                us_env=True,
                                                ignore_returncodes=[127])
        if value == 0:
            return output
        raise UnilintPluginInitException("ERROR Cannot find pylint, install via apt or pip\n%s"%message)

    def check_resource(self, options, path, type_categories):
        if (not 'python-src' in type_categories and
            not 'script' in type_categories):
            return None

        levels = []

        if options.level > LEVELS[LEVEL_WARNING]:
            levels.append('w')  # warn
        if options.level > LEVELS[LEVEL_STYLE]:
            levels.append('r')  # refactor
            levels.append('c')  # convention

        lvl_cmd = ''
        if levels != []:
            lvl_cmd = "-d%s"%','.join(levels)

        suppress_options = []
        if 'script' in type_categories:
            suppress_options.append('-d E0611')
        cmd = 'pylint %s --output-format=parseable -i y -r n %s %s'%(path, lvl_cmd, ' '.join(suppress_options))
        if options.verbose:
            print(cmd)
        _, output, _ = self.shell_cmd(cmd,
                                      shell=True,
                                      us_env=True,
                                      cwd=os.path.dirname(path),
                                      ignore_returncodes=range(50))
        output += '\n'
        basepath = os.path.dirname(path)
        if options.raw:
            print output
            return None
        return self.parse(output, basepath)

    def parse(self, output, basepath):
        # Examle output:
        #         dotcode_tf.py:39: [C0111, RosTfTreeDotcodeGenerator] Missing docstring
        #         dotcode_tf.py:47: [C0322, RosTfTreeDotcodeGenerator.__init__] Operator not preceded by a space
        #         self.listen_duration=1
        #                             ^
        issues = []

        lines = []
        # split lines except for multiline messages
        for line in output.splitlines():
            if line.strip() == '':
                continue
            if NEWLINE_MATCHER.match(line) or len(lines) == 0:
                lines.append(line)
            else:
                lines[-1] += "\n%s"%line

        for line in lines:
            issue_lines = line.splitlines()
            message_line = issue_lines[0]
            if len(issue_lines) > 1:
                # source = issue_lines[1]
                position = len(issue_lines[2]) -1
            else:
                # source = None
                position = None

            filename, line_number, rest = message_line.split(':', 2)

            category, message = rest.lstrip('[ ').split(']', 1)
            if ' ' in category:
                category = category.split()[0]

            severity = None
            if category.startswith('E'):
                severity = "error"
            if category.startswith('F'):
                severity = "error"
            if category.startswith('C'):
                severity = "style"
            if category.startswith('R'):
                severity = "warning"
            if category.startswith('W'):
                severity = "warning"

            issue = Issue(path=os.path.join(basepath, filename),
                          message=message.strip(),
                          checker_id=self.get_id(),
                          line_number_start=line_number,
                          line_number_end=line_number,
                          line_position=position,
                          severity=severity,
                          category=category.rstrip(','))
            issues.append(issue)
        return issues
